"use strict";
// -*- mode: typescript; indent-tabs-mode: nil; js-basic-offset: 4 -*-
//
// This file is part of Genie
//
// Copyright 2020-2021 The Board of Trustees of the Leland Stanford Junior University
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//    http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: Giovanni Campagna <gcampagn@cs.stanford.edu>
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const sql = __importStar(require("."));
class SyncTable {
    constructor(db, name, fields) {
        this.name = name;
        this.fields = fields;
        this._db = db;
        this._discriminator = fields[0];
    }
    _getLastModifiedInternal(client) {
        return sql.selectAll(client, `select max(lastModified) as maxLastModified
                             from ${this.name}_journal`, []).then((rows) => {
            if (rows.length === 0 || rows[0].maxLastModified === null)
                return 0;
            else
                return rows[0].maxLastModified;
        });
    }
    getAll() {
        return this._db.withClient((client) => {
            return sql.selectAll(client, `select uniqueId, ${this.fields} from ${this.name}`, []);
        });
    }
    getOne(uniqueId) {
        return this._db.withClient((client) => {
            return sql.selectOne(client, `select uniqueId,${this.fields.join(',')}
                                 from ${this.name} where uniqueId = ?`, [uniqueId]);
        });
    }
    getRaw() {
        return this._db.withClient((client) => {
            return sql.selectAll(client, `select tj.uniqueId,tj.lastModified,${this.fields.map((f) => 't.' + f)}
                                 from ${this.name}_journal as tj left outer join
                                 ${this.name} as t on tj.uniqueId = t.uniqueId`, []);
        });
    }
    getChangesAfter(lastModified) {
        return this._db.withClient((client) => {
            return sql.selectAll(client, `select tj.uniqueId,tj.lastModified,${this.fields.map((f) => 't.' + f).join(',')}
                                 from ${this.name}_journal as tj left outer join
                                 ${this.name} as t on tj.uniqueId = t.uniqueId where
                                 tj.lastModified > ?`, [lastModified]);
        });
    }
    _handleChangesInternal(client, changes) {
        return Promise.all(changes.map((change) => {
            if (change[this._discriminator] !== null) {
                return this._insertIfRecentInternal(client, change.uniqueId, change.lastModified, change);
            }
            else {
                return this._deleteIfRecentInternal(client, change.uniqueId, change.lastModified);
            }
        }));
    }
    handleChanges(changes) {
        return this._db.withTransaction((client) => {
            return this._handleChangesInternal(client, changes);
        });
    }
    syncAt(theirLastModified, pushedChanges) {
        return this._db.withTransaction(async (client) => {
            const ourChanges = await sql.selectAll(client, `select tj.uniqueId,tj.lastModified,${this.fields.map((f) => 't.' + f)}
                 from ${this.name}_journal as tj left outer join
                 ${this.name} as t on tj.uniqueId = t.uniqueId where
                 tj.lastModified > ?`, [theirLastModified]);
            const lastModified = await this._getLastModifiedInternal(client);
            const done = await this._handleChangesInternal(client, pushedChanges);
            return { lastModified, ourChanges, done };
        });
    }
    async _insertInternal(client, uniqueId, lastModified, row) {
        const insertSql = `insert or replace into ${this.name} (uniqueId,${this.fields})
            values(?,${this.fields.map(() => '?')})`;
        const param = [uniqueId].concat(this.fields.map((f) => row[f]));
        await sql.insertOne(client, insertSql, param);
        await sql.insertOne(client, `insert or replace into ${this.name}_journal
            (uniqueId, lastModified) values(?, ?)`, [uniqueId, lastModified]);
        return lastModified;
    }
    replaceAll(data) {
        return this._db.withTransaction(async (client) => {
            await sql.query(client, `delete from ${this.name}`, []);
            await sql.query(client, `delete from ${this.name}_journal`, []);
            await Promise.all(data.map(async (row) => {
                const uniqueId = row.uniqueId;
                const lastModified = row.lastModified;
                if (row[this._discriminator] === null)
                    return;
                await this._insertInternal(client, uniqueId, lastModified, row);
            }));
        });
    }
    _insertIfRecentInternal(client, uniqueId, lastModified, row) {
        return sql.selectAll(client, `select lastModified from ${this.name}_journal where uniqueId = ?`, [uniqueId]).then((rows) => {
            if (rows.length > 0 && rows[0].lastModified >= lastModified)
                return false;
            return this._insertInternal(client, uniqueId, lastModified, row).then(() => true);
        });
    }
    insertIfRecent(uniqueId, lastModified, row) {
        return this._db.withTransaction((client) => {
            return this._insertIfRecentInternal(client, uniqueId, lastModified, row);
        });
    }
    insertOne(uniqueId, row) {
        return this._db.withTransaction((client) => {
            const now = (new Date).getTime();
            return this._insertInternal(client, uniqueId, now, row);
        });
    }
    async _deleteInternal(client, uniqueId, lastModified) {
        await sql.insertOne(client, `delete from ${this.name} where uniqueId = ?`, [uniqueId]);
        await sql.insertOne(client, `insert or replace into ${this.name}_journal
                            (uniqueId, lastModified) values(?, ?)`, [uniqueId, lastModified]);
        return lastModified;
    }
    _deleteIfRecentInternal(client, uniqueId, lastModified) {
        return sql.selectAll(client, `select lastModified from ${this.name}_journal where uniqueId = ?`, [uniqueId]).then((rows) => {
            if (rows.length > 0 && rows[0].lastModified >= lastModified)
                return false;
            return this._deleteInternal(client, uniqueId, lastModified).then(() => true);
        });
    }
    deleteIfRecent(uniqueId, lastModified) {
        return this._db.withTransaction((client) => {
            return this._deleteIfRecentInternal(client, uniqueId, lastModified);
        });
    }
    deleteOne(uniqueId) {
        return this._db.withTransaction((client) => {
            return this._deleteInternal(client, uniqueId, Date.now());
        });
    }
}
exports.default = SyncTable;
//# sourceMappingURL=sync_table.js.map